/* ICache.H
 * General Icache class
 */

inline
int
Log2(ulong Value)
{
  int i;
  for (i=0; Value>>i; i++);
  return i-1;
}


class ICache {
private:
  int Height;
  int Width;
  int Associativity;
  int Size;

  ushort LineBits;
  ushort TagShift;

  // Used to do LRU/NMRU
  unsigned long Cycle;

  // unsigned int *Data;
  unsigned long *Tags;
  unsigned short *Valid;
  unsigned short *Owner;
  unsigned long *TimeStamp;

  int Refs;
  int Hits;
  /* other variables for maintaining stats go here */

public:
  // width in bytes; associativity in # ways
  ICache(int height, int width, int associativity)
  {
    Height = height;
    Width = width;
    Associativity = associativity;

    Size = height*width;

    // Data = new unsigned int [Size];
    Tags = new unsigned long [Height * Associativity];
    Valid = new unsigned short [Height * Associativity];
    TimeStamp = new unsigned long [Height * Associativity];

    LineBits = Log2(Width/Associativity);
    TagShift = Log2(Height) + LineBits;

    Refs = 0;
    Hits = 0;
    Cycle = 0;
  }

  // Don't replace if missing
  int Lookup(unsigned long Addr, int who)
  {
    // Return true iff hit
    uint line = (Addr >> LineBits) & (Height-1);
    uint LiAs = (line*Associativity);
    ulong tag = (Addr >> TagShift);
    int way;

    Refs++;
    Cycle++;

    for (way=0; way<Associativity; way++) {
      if ((Tags[LiAs+way]==tag) && (Valid[LiAs+way])) {
        TimeStamp[LiAs+way] = Cycle-1;
        Hits++;
        return 1;
      }
    }
    return 0;
  }

  // Regular cache reference
  int Fetch(unsigned long Addr, int who)
  {
    // Return true iff hit

    uint line = (Addr >> LineBits) & (Height-1);
    uint LiAs = line*Associativity;
    ulong tag = (Addr >> TagShift);
    int way, MinWay;
    ulong MinTime;

    Refs++;
    Cycle++;

    for (way=0; way<Associativity; way++) {
      if ((Tags[LiAs+way]==tag) && Valid[LiAs+way]){
        TimeStamp[LiAs+way] = Cycle-1;
        Hits++;
        return 1;
      }
    }

    // This was a miss. Read in line from next level
 
    // Find LRU line
    MinTime = (~(1ul))>>1;
    for (way=0; way<Associativity; way++) {
      ulong t;
      if ((t=TimeStamp[LiAs+way]) < MinTime) {
        MinWay = way;
        MinTime = t;
        if (t==0) break; // t==0 is equivalent to valid==0
      }
    }

    // Replace this line
    Tags[LiAs+MinWay] = tag;
    TimeStamp[LiAs+MinWay] = Cycle-1;
    Valid[LiAs+MinWay] = 1;

    return 0;
  }

  // Invalidate a cacheline
  void Invalidate(unsigned long Addr);

  void PrintStats()
  {
    cerr << endl
         << "ICache statistics:" << endl << endl
         << "Refs=" << Refs
         << ", Hits="<< Hits << endl
         << "Hit-Rate=" << ((float)Hits)/((float)Refs)
         << endl;
  }
};

