

// btb.C: a simple btb/icache simulator

#include <iostream.h>

#include <aint.h>

// Icache: 256 lines, 32 bytes (8 insts) per line
const int CacheLines = 256;
const int CacheWidth = 32;

class cache_line {
public:
  cache_line() {valid=0;}
  ulong Tag;
  int valid;
  int Ref(ulong tag) { int hit = ( (Tag==tag) && (valid)); Tag = tag; valid = 1; return hit; }
};

class Icache {
public:
  Icache(int lines=CacheLines, int width=CacheWidth)
  {
    Lines = lines;
    Width = width;
    Size = lines*width;
    Line = new cache_line[Lines];
    refs = hits = 0;
  }
  cache_line *Line;
  int Lines;
  int Width;
  int Size;

  int refs;
  int hits;

  void Fetch(aint_addr_t pc)
  {
    int line = (pc/Width) & (Lines-1);
    ulong tag = pc/Size;
    if (Line[line].Ref(tag)) hits++;
    refs++;
  }

  aint_addr_t Build(ushort line, uchar offset)
  {
    return (Line[line].Tag*Size + line*Width + offset);
  }

  void PrintStats()
  {
    cerr << endl << "ICache Stats: " << endl
	 << "Refs = " << refs << endl
	 << "Hits = " << hits << endl
	 << "Miss Rate = " << 1.0 - ( (float) hits / (float) refs ) << endl;
  }
};

// BTB: 4k lines; no tag; cache-line+offset in data
const int BTB_Size = (1<<12);

class btb_line {
public:
  btb_line() {valid=0; offset=0; cacheline=0;}
  uchar valid;
  uchar offset;     // offset of target inst in cacheline
  ushort cacheline; // target cacheline
    
};

class BTB {
private:
  int Size;
  btb_line *Line;
  int refs;
  int hits;
  Icache *IC;
  ulong last_target;
  ulong last_pc;
public:
  BTB(int size=BTB_Size) { Size=size; Line = new btb_line[Size]; refs=0; hits=0; IC = new Icache(CacheLines, CacheWidth);}

  void Fetch(aint_addr_t pc, int is_condbr)
  {
    if (is_condbr) {
      int line=((pc>>2) & (Size-1));
      last_pc = pc;
      last_target = IC->Build(Line[line].cacheline, Line[line].offset);  // dont worry about valid
      refs++;
    } else {
      IC->Fetch(pc);
    }
  }

  void Verify(aint_addr_t target)
  {
    if (last_target==target) hits++;
    else {
      int line=((last_pc>>2) & (Size-1));
      Line[line].offset = target&(CacheWidth-1);
      Line[line].cacheline = ((target/CacheWidth) & (CacheLines-1));
      Line[line].valid=1;
      IC->Fetch(target);
    }
  }

  void PrintStats()
  {
    cerr << endl << "BTB and ICache Stats" << endl;
    IC->PrintStats();
    cerr << endl << "BTB Stats: " << endl;
    cerr << "Refs = " << refs << endl
	 << "Hits = " << hits << endl
	 << "BTB Miss Rate = " << 1.0 - ( (float) hits / (float) refs ) << endl;
      
  }
};

aint_addr_t PC;    // address to fetch from
path_ptr ThePath;  // Main path in the main thread in the target binary
BTB TheBTB;



// Aint calls this function before opening any object files - use this
// to initialize back-end variables
int
sim_init(int argc, char **argv, char **envp)
{
  /* AINT uses the getopt() function to parse the argument string, and
   * expects optind to have the correct value (first object file name)
   * upon return from this function. We must therefore use getopt() to
   * process arguments meant for us (delimited by "--").
   */

  extern int optind;    // AINT uses the getopt() function to keep track of args

  // For this example, we use no arguments...

  // Return optind so that AINT finds the name of the target binary...
  return optind;
}


// Aint calls this function if the command line has an error
void
sim_usage(int argc, char **argv)
{
  cerr << endl << "Usage: <simname> [AINT options] executable" << endl;
  cerr << "This simulator takes no arguments, and accepts one executable" << endl;
}


// AINT calls this function when it creates the main thread in the process
// corresponding to the target binary
void
sim_thread_begin(int tid, aint_addr_t pc)
{
  thread_ptr pthread = Threads[tid];    // Get thread pointer from AINT's table
  ThePath = AINT_create_path(tid, pc);    // Ask AINT to create main path

  informative("sim_thread_begin: began thread/path" << pthread->tid
      << " " << (void *) pthread->path_tree->pathid << endl);

  PC = pc;    // Starting address for thread
}


// The main execution loop - fetch, execute, commit each inst one at a time
void
sim_execute()
{
  inflight_inst_ptr ifi;

  informative("sim_execute: commencing execution at " << (void *) PC << endl);

  while(PC) {    // we get a zero PC in *ifi when program executes _exit()

    ifi = AINT_fetch(ThePath, PC);    // fetch next instruction
    informative("Fetch: " << (void *) PC << endl);
    TheBTB.Fetch(PC, (ifi->iflags & E_CONDBR));

    AINT_issue(ifi);    // ask AINT to issue the fetched instruction
    informative("Issue: " << (void *) PC << endl);

    if (ifi->iflags & E_READ) {    // this is a load inst
      AINT_do_read(ifi);
    } else if (ifi->iflags & E_WRITE) {    // this is a store inst
      AINT_do_write(ifi);
    }

    PC = ifi->nextPC;    // AINT updates ifi->nextPC upon executing inst

    if (ifi->iflags & E_CONDBR) TheBTB.Verify(PC);
    AINT_commit(ifi);    // Done
  }

}  // sim_execute()


void
sim_done()
{
  TheBTB.PrintStats();
}
