/* tlb.h
 * structures for the paged data and instruction spaces
 */

#ifndef AINT_TLB_H
#define AINT_TLB_H

/* Page-Table entry for the per-process TB */
class segment;
class tb_entry {
public:
  tb_entry() {
    next = NULL; page = NULL; textpage = NULL; lookaside = NULL;
    tag = 0; flags = 0; prot = 0; BackEndRefCount = 0;
  }
  tb_entry *next;
  void *page;
  unsigned long *fe_vector;    /* full-empty bits */
  icode_ptr textpage; /* for text pages, contains decoded text */
  tb_entry *lookaside; /* Used only at the head of the queue;
				 also pads up to power of 2 */
  aint_addr_t vaddr; // base vaddr of this page
  class segment *segment;

  ulong tag;
  int flags;                  /* Holds MAP_xxx flags, including MAP_PRIVATE/MAP_SHARED flag */
  int prot;                   /* Holds PROT_xxx flags */

  long BackEndRefCount;    // NOT maintained by AINT - for backend use
  void BackEndReference() { BackEndRefCount++; }
  long GetRefCount() { return BackEndRefCount; }
};

/* Constants and macros for the per-process Translation Buffer */
const int TB_OFFSET_LENGTH = 13;
const int TB_PAGESIZE = (1 << TB_OFFSET_LENGTH);
const int TB_KEY_LENGTH = 6;
const int TB_SIZE = (1 << TB_KEY_LENGTH);
/* full-empty bits. there is 1 bit per 64-bit word; 1 byte per 64 bytes */
const int TB_FE_VEC_LENGTH = (TB_OFFSET_LENGTH-6); /* length of fe vec addr */
const int TB_FE_VEC_SIZE = (1 << TB_FE_VEC_LENGTH); /* in bytes */
const int TB_FE_VEC_LONGS = (TB_FE_VEC_SIZE << 3);      /* in 8-byte quadwords */

inline long TB_VADDR(aint_addr_t vaddr) {
  return (vaddr & ~((long)TB_PAGESIZE-1));
}

inline long TB_VPAGE(aint_addr_t vaddr) {
  return vaddr >> TB_OFFSET_LENGTH;
}

inline long TB_OFFSET(aint_addr_t vaddr) {
  return vaddr & (TB_PAGESIZE-1);
}

inline int TB_KEY(long vpage) {
  return vpage & (TB_SIZE - 1);
}

inline long TB_TAG(long vpage) {
  return vpage >> TB_KEY_LENGTH;
}


class TLB {
public:
  TLB(process_ptr myprocess) { pprocess = myprocess; for(int i=0; i<TB_SIZE; i++) TB[i] = (tb_entry *) NULL; }

  tb_entry *Translate(aint_addr_t addr, long *paccess);
  tb_entry *miss(aint_addr_t addr, long *paccess);

protected:

  tb_entry *PageList(ulong _key) { return TB[_key]; }
  tb_entry *Lookaside(ulong _key) { return TB[_key]->lookaside; }
  void Lookaside(ulong _key, tb_entry *_pg) { TB[_key]->lookaside =_pg; }

  tb_entry *TB[TB_SIZE];
  process_ptr pprocess;

};

/* returns instruction number */
inline ulong ITB_OFFSET(aint_addr_t vaddr) { return ((vaddr & (TB_PAGESIZE-1)) >> 2); }

// Return pointer to tb entry if address can be found; else return null
inline tb_entry *
TLB::Translate(aint_addr_t addr, long *paccess)
{
  ulong vpage, tbkey, tbtag;
  tb_entry *pg;

  vpage = TB_VPAGE(addr);
  tbkey = TB_KEY(vpage);
  tbtag = TB_TAG(vpage);

  pg = PageList(tbkey);
  if (pg) {
    /* first page on process's TB[] list must have a lookaside value */
    // assert(pg->lookaside);
    if (pg->lookaside->tag == tbtag) {
      pg = pg->lookaside;
    } else {
      for (; pg; pg = pg->next) {
	if (pg->tag == tbtag) { 
	  /* TB HIT */
	  /* If we are here, then pg is non-zero, ie TB[tbkey] exists */
	  Lookaside(tbkey, pg);
	  break;
	}
      }
    }
  }

  if (pg) {
    if (paccess) {
      *paccess = pg->lookaside->prot;
    }
  } else {
    pg = miss(addr, paccess);
  }

  return pg;
}


#endif
