/* storebuf.c 
 * routines for manipulating the store buffer
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>

#include "aint.h"
#include <globals.h>
#include <opcodes.h>
#include <protos.h>


/* This routine updates the store buffer entry using the vaddr of
 * the store. The actual updates and byte masks are in the ifi struct.
 *
 * pthread->StoreQ is the head of the address-entry doubly-linked list.
 * Each sqe_addr contains the head of the updates list
 * The updates list is sequential; updates are inserted in fetch order
 */

void
Store_Buffer::Update(inflight_inst_ptr ifi)
{
  struct stb_addr *addr_elem;
  struct stb_entry *update;

  /* Look up the storeq for existing updates to this address */
  QSEARCH(StoreQ, (ifi->vaddr & QWORD_MASK), addr_elem);

  if (addr_elem == StoreQ) {
    /* no pending updates to this address */
    addr_elem = new stb_addr(StoreQ, (ifi->vaddr & QWORD_MASK));

    /* Insert update into list for this addr */
    ifi->stb_update = new stb_entry(addr_elem->updates, ifi);
    
    return;
  }

  /* We have an existing storeq address entry. Insert this one in fetch order */
  for (update=addr_elem->updates->next; (update != addr_elem->updates); update=update->next) {
    if (update->fetchnum > ifi->fetchnum) {
      /* insert ourself before sqe since they come after us */
      ifi->stb_update = new stb_entry(update->prev, ifi);
      return;
    } else if (update->fetchnum == ifi->fetchnum) {
      /* Overwrite our own prior update */
      ifi->stb_update = new stb_entry(update, ifi);
      delete update;
      return;
    }
  }

  /* We reached the end of the addr list */
  if (update->next == addr_elem->updates) fatal("Update_StoreBuffer: found empty update list");
      
  /* Enqueue us here since we are the highest fetchnum */
  ifi->stb_update = new stb_entry(update->prev, ifi);
}



/* For store insts, remove pending updates from store queue */
void
Store_Buffer::Remove(inflight_inst_ptr ifi)
{
  struct stb_addr *addr_elem;

  if (ifi->stb_update) {
    if (ifi->stb_update->next == NULL) 
      warning("(ifi->stb_update->next == NULL)");
    if (ifi->stb_update->prev == NULL) 
      warning("(ifi->stb_update->prev == NULL)");

    if (ifi->stb_update->next == ifi->stb_update->prev) {
      /* This was the last remaining update to this address */
      /* Find the stq addr entry for this update */
      QSEARCH(StoreQ, (ifi->vaddr & QWORD_MASK), addr_elem);
      if (addr_elem == StoreQ) 
	fatal("Cleanup_StoreBuffer: address " << (void *) ifi->vaddr << " not found in store-buffer" << endl);
      delete(ifi->stb_update);
      delete(addr_elem);
      // ifi->stb_update = NULL;
      return;
    } 
    delete (ifi->stb_update);
    // ifi->stb_update = NULL;
  }
}


void Store_Buffer::Lookup(inflight_inst_ptr ifi, ulong &bitmask)
{
  /* Search the storeq for most recent updates */
  struct stb_addr *addr_elem;
  struct stb_entry *update;
  ulong tempmask;
  inflight_inst_ptr u_ifi;

  QSEARCH(StoreQ, (ifi->vaddr & QWORD_MASK), addr_elem);
  if (addr_elem == StoreQ) return;    // no updates in storebuffer

  for (update = addr_elem->updates->prev; update != addr_elem->updates; update = update->prev) {
    /* Do not consider updates from newer stores and non-ancestor paths*/
    if (update->fetchnum > ifi->fetchnum) continue;

    u_ifi = update->ifi;

    if (!IS_ANCESTOR(u_ifi->ppath->pathid, ifi->ppath->pathid)) continue;
    
    /* This is an update from an older store */
    tempmask = bitmask & (u_ifi->bitmask);
    /* tempmask contains the valid bits from this update */

    /* u_ifi->data is the actual (unshifted) word written at the actual (un-realigned) address.
     * We need to convert it to a qword-aligned data word
     */
    ifi->data &= (~tempmask); // zero out bits to read from u_ifi->data
    ifi->data |= (tempmask & (u_ifi->data << ((u_ifi->vaddr & 0x7)<<3)));

    bitmask &= (~tempmask);

  }
}
